\ww.xpl		16-Jul-2017
\Lorenzian Waterwheel animated using page flipping
\
\REVISIONS
\23-Oct-2003: Original for PC.
\16-Jul-2017: Extensively modified for RPi.

def	Buckets = 8;	\number of buckets in waterwheel
def 	BWidth=50, BHeight=55, B2Width=BWidth/2;  \bucket dimensions (pixels)
def	Base = 480-70;	\Y coordinate for baseline (X axis) of speed graph
def	SLimit = 57;	\limit for speed graph (so it doesn't mess up screen)

real	Speed,	\speed of waterwheel rotation (radians per iteration)
	Fric,	\friction slowdown rate (percent slowdown per iteration)
	WaterRate, \water fill rate (arbitrary units)
	Leak,	\bucket leak rate (percent leaked per iteration)
	Ang0,	\angle of reference point (bucket 0) on wheel (radians)
	AngB,	\angle between buckets around wheel (radians)
	Torque,	\torque on wheel caused by weight of water in buckets
	Water(Buckets),	\amount of water in each bucket
	BucX(Buckets),	\X coordinate of center of bucket
	BucY(Buckets),	\Y coordinate of top of bucket	   (0,0 = center)
	Ang(Buckets);	\angle of bucket around waterwheel (radians)

int	Page,	\video page being drawn onto (0 or 1)
	B,	\bucket number (0..Buckets-1)
	BX, BY,	\coordinates for a bucket (pixels)
	SX, SY,	\coordinates for plotting speed graph (pixels)
	SX0,	\steps per pixel for SX
	WLev;	\depth of water level in a bucket (pixels)

def	Bel=$07, Esc=$1B, Sp=$20, Ctrl=$40;	\ASCII control characters
def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,	\attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \ EGA palette
def	Pi = 3.14159265358979323846, Pi2 = Pi*2.0;



func	Ror(N, C);		\Rotate N right C bits
int	N, C;
def	M = $000F;		\mask to rotate within 16 bits
return N>>(C&M) ! N<<(16-(C&M));



proc	CTxt2(X, Y, Str);	\Display a text string at cursor position X, Y
int	X, Y;			\This displays on both video pages
char	Str;
[Cursor(X, Y);  Text(6, Str);
Cursor(X, Y+30);  Text(6, Str);
];	\CTxt2



proc	CRlOut2(X, Y, Num);	\Display a real number at cursor position X, Y
int	X, Y;			\This displays on both video pages
real	Num;
[Cursor(X, Y);  RlOut(6, Num);
Cursor(X, Y+30);  RlOut(6, Num);
];	\CRlOut2



proc	DrawLine(X0, Y0, X1, Y1, C);	\Draw a line segment
\Draw onto the portion of video memory that is not currently being viewed
int	X0, Y0, X1, Y1, C;	\end coordiantes, color
def	H=-5-16, V=40;		\centering adjustment (8 pixels per char cell)
[Move(X0+H, Y0+V+Page*480);
 Line(X1+H, Y1+V+Page*480, C);
];	\DrawLine



proc	DrawBlock(X0, Y0, X1, Y1, C);	\Draw a filled rectangle
int	X0, Y0, X1, Y1, C;	\coordinates of opposite corners, fill color
int	I, J, T;
[if X1 < X0 then [T:=X1;  X1:=X0;  X0:=T];	\put coordinates into
 if Y1 < Y0 then [T:=Y1;  Y1:=Y0;  Y0:=T];	\ ascending order

if (Y1-Y0) > 8*(X1-X0) then		\if tall and skinny then
	for I:= X0 to X1 do		\ draw vertical lines
		DrawLine(I, Y0, I, Y1, C)
else	for J:= Y0 to Y1 do		\ else draw fast horizontal lines
		DrawLine(X0, J, X1, J, C);
];	\DrawBlock



proc	DrawBucket(X0, Y0, X1, Y1, C);	\Draw a bucket
int	X0, Y0,		\coordinate of upper-left corner
	X1, Y1,		\coordiante of lower-right corner
	C;		\color
[DrawLine(X0, Y0, X0, Y1, C);
 DrawLine(X0, Y1, X1, Y1, C);
 DrawLine(X1, Y1, X1, Y0, C);
];	\DrawBucket

\-------------------------------------------------------------------------------

[TrapC(true);	\(prevent Ctrl-C'ing out of a program in graphic mode)
Format(3,1);	\display format of floating point numbers = XXX.X

\Set initial conditions
Speed:= 0.04;
WaterRate:= 20.0;
Fric:= 0.02;
Leak:= 0.006;
for B:= 0 to Buckets-1 do Water(B):= 0.0;	\no water in buckets
Ang0:= 0.0;	\angle of reference point on wheel for bucket 0
AngB:= Pi2 / float(Buckets);	\angle between buckets
SX:= 0;		\start speed graph at left edge of screen
SX0:= 0;

SetVid($12);	\set display for 640x480 graphics
Page:= 1;	\viewing page 0; draw on page 1

Attrib(LCyan);
CTxt2(28, 0, "LORENZIAN WATERWHEEL");

Attrib(White);
CTxt2( 2, 2, "[S]peed:     ");
CTxt2( 2, 3, "[F]riction:       %");
CTxt2( 2, 4, "[W]ater Rate:");
CTxt2( 2, 5, "[L]eak Rate:      %");
CTxt2(54, 2, "Uppercase increases value");
CTxt2(54, 3, "Lowercase decreases value");
CTxt2(54, 4, "Spacebar empties buckets");
CTxt2(54, 5, "Esc terminates program");

Attrib(Cyan);
CTxt2( 2, 19, "Speed:");			\label for graph

Move(0, Base);  Line(639, Base, Cyan);		\draw baseline of speed graph
Move(0, Base+480);  Line(639, Base+480, Cyan);	\ on both video pages

loop	[Attrib(White);
	CRlOut2(15, 2, Speed * 180.0/Pi);	\degrees/unit_time
	CRlOut2(15, 3, Fric*100.0);		\percent of slowdown
	CRlOut2(15, 4, WaterRate);		\arbitrary units
	CRlOut2(15, 5, Leak*100.0);		\percent of water in bucket

	Ang0:= Ang0 + Speed;		\rotate wheel as a function of speed
	while Ang0 < 0.0 do Ang0:= Ang0 + Pi2;	\limit angle to one revolution
	while Ang0 > Pi2 do Ang0:= Ang0 - Pi2;

	Torque:= 0.0;				\initialize torque accumulator
	for B:= 0 to Buckets-1 do		\for all of the buckets...
		[Ang(B):= Ang0 + AngB*float(B);	\bucket angle around wheel
		BucX(B):= Sin(Ang(B)) * 100.0;	\bucket coordinates
		BucY(B):= Cos(Ang(B)) * 100.0;

		\Fill bucket under faucet
		if abs(BucX(B))<=float(B2Width) & BucY(B)>0.0 then
			Water(B):= Water(B) + WaterRate;

		\Leak water out (rate is proportional to amount of water)
		Water(B):= Water(B) * (1.0-Leak);

		WLev:= fix(Water(B)/float(BWidth));	\limit amount of water
		if WLev > BHeight then Water(B):= float(BHeight*BWidth);
		if WLev < 0 then Water(B):= 0.0; \(negative water is ugly!)

		\Sum torques (proportional to amount of water * X coordinate)
		Torque:= Torque + Water(B) * BucX(B);
		];
	\Change in speed is proportional to torque
	Speed:= Speed + Torque/80E6;
	\(Experimentally determined fudge factor; it keeps values reasonable)

	\Reduce speed as a function of friction
	Speed:= Speed * (1.0-Fric);
	\(Linear friction (subtraction) doesn't cause chaos; this is more like
	\ air friction)

	\Erase old wheel image
	DrawBlock(320-100-B2Width, 0, 320+100+B2Width, 200+20+BHeight, Black);

	\Draw a picture of the results
	Attrib(White);
	CTxt2(37, 2, "o");		\faucet

	Attrib(Brown);
	CTxt2(37, 11, "o");		\axle

	\Draw water spilling from faucet
	DrawBlock(320-1, 3, 320+1, 75, Ror($63,SX)<<8 +LBlue);

	\Draw buckets
	for B:= 0 to Buckets-1 do
		[BX:= fix(BucX(B)) + 320;
		BY:= 100 - fix(BucY(B)) + 20;
		WLev:= fix(Water(B)/float(BWidth));
		DrawBlock(BX-B2Width, BY+BHeight, BX+B2Width, BY+BHeight-WLev,
			LBlue);
		DrawBucket(BX-B2Width, BY, BX+B2Width, BY+BHeight, 
			if B=0 then Yellow else Brown);
		];

	\Draw graph of speed vs. time
	Move(SX, Base-SLimit+Page*480);		\erase old value
	Line(SX, Base+SLimit+Page*480, Black);
	Point(SX, Base+Page*480, Cyan);		\restore baseline

	SY:= fix(Speed*500.0);			\coordinate for speed
	if SY > SLimit then SY:= SLimit;	\ (amplitude is arbitrary)
	if SY < -SLimit then SY:= -SLimit;	\ limits prevent screen glitches
	Point(SX, Base-SY+Page*480, LGreen);	\make clockwise speed positive

	Move(SX+1, Base-SLimit+Page*480);	\draw vertical graph line
	Line(SX+1, Base+SLimit+Page*480, Cyan);

	SX0:= SX0+1;				\plot 4 pixels per X position
	if SX0 = 4 then
		[SX0:= 0;
		SX:= SX+1;			\advance to next point
		];
	if SX >= 640 then SX:= 0;		\wrap around screen

	ShowPage(Page);				\show what was drawn
	WaitForVSync;				\ at the next video frame
	Page:= Page | 1;			\flip to page not being shown

	if KeyHit then				\handle keyboard commands
		[case ChIn(1) of
		  ^S:	Speed:= Speed + 0.01;	\speed adjustment
		  ^s:	Speed:= Speed - 0.01;	\ (sometimes it needs a nudge)
		  ^F:	Fric:= Fric + 0.001;	\friction adjustment
		  ^f:	Fric:= Fric - 0.001;
		  ^W:	WaterRate:= WaterRate + 0.1;
		  ^w:	WaterRate:= WaterRate - 0.1;
		  ^L:	Leak:= Leak + 0.001;
		  ^l:	Leak:= Leak - 0.001;
		  Sp:	for B:= 0 to Buckets-1 do Water(B):= 0.0; \empty buckets
		  ^C-Ctrl, Esc:	quit		\terminate program
		other	ChOut(0, Bel);
		];
	];	\loop
SetVid($03);					\restore normal text display
]
